@@ -1,4 +1,4 @@
-# ------------------ Safetensors2 1.0 by DeepBeepMeep (mmgp)------------------
+# ------------------ Safetensors2 1.3 by DeepBeepMeep (mmgp)------------------
 #
 # This module entirely written in Python is a replacement for the safetensor library which requires much less RAM to load models.
 # It can be conveniently used to keep a low RAM consumption when handling  transit data (for instance when quantizing or transferring tensors to reserver RAM)
@@ -16,12 +16,14 @@ import safetensors
 import accelerate
 import os
 from collections import OrderedDict
+import warnings
 
+warnings.filterwarnings("ignore", ".*The given buffer is not writable, and PyTorch does not support non-writable tensors*")
 
 _old_torch_load_file = None
 _old_safe_open = None
 
-
+all_tensors_are_read_only = False 
 
 mmm = {}
 verboseLevel = 1
@@ -44,7 +46,16 @@ class MmapTracker:
         file_path = os.path.join(*s)
         self.file_path = file_path # os.path.abspath(file_path) 
         self.count = 0
-        mmm[file_path] = self
+        key = file_path
+        i = 1
+        while True:
+            if key not in mmm:
+                mmm[key] = self
+                break
+            i +=1
+            key = key + "#" + str(i)
+        self.mmm_key = key
+        # print(f"MMAP Add: {file_path}: {mmm.keys()}")
 
     def register(self, mmap_obj, map_id, start, size):
 
@@ -59,7 +70,8 @@ class MmapTracker:
 
                 print(f"MMap Manager of file '{self.file_path}' : MMap no {map_id} has been released" + text)
             if self.count == self._already_released:
-                del mmm[self.file_path]
+                # print(f"MMAP Del: {self.file_path}: {mmm.keys()}")
+                del mmm[self.mmm_key ]
 
             self._maps.pop(map_id, None)
 
@@ -75,6 +87,24 @@ class MmapTracker:
     def get_active_maps(self):
         return dict(self._maps)
 
+class tensor_slice:
+    catalog = None
+    value = None
+    name = None
+
+    def __init__(self, catalog, name, value):
+        self.catalog = catalog
+        self.value = value
+        self.name = name
+
+    def __getitem__(self, s):
+        return self.value[s]
+ 
+    def get_dtype(self):
+        return self.catalog[self.name]["dtype"]
+
+    def get_shape(self):
+        return self.catalog[self.name]["shape"]
 
 class cached_metadata:
     file_path = None
@@ -105,20 +135,18 @@ class cached_metadata:
 _cached_entry = None # ideally we should create a dict of the last n entries but one entry covers most cases
 
 def  _parse_metadata(metadata):
-    if metadata == None:
-        return None
-    
     new_metadata= {}
-    
-    for k,v in metadata.items():
-        if k.endswith("_base64"):
-            v_decoded = json.loads(base64.b64decode(v.encode('utf8')).decode('utf8'))
-            p = k.rfind("_")
-            new_k = k[:p]
-            new_metadata[new_k]= v_decoded
-        else:
-            new_metadata[k] = v
-
+    if metadata != None:
+        for k,v in metadata.items():
+            if k.endswith("_base64"):
+                v_decoded = json.loads(base64.b64decode(v.encode('utf8')).decode('utf8'))
+                p = k.rfind("_")
+                new_k = k[:p]
+                new_metadata[new_k]= v_decoded
+            else:
+                new_metadata[k] = v
+    if "format" not in new_metadata:
+        new_metadata["format"] = "pt"
     return new_metadata
 
 def _read_safetensors_header(path, file):
@@ -146,7 +174,7 @@ def _read_safetensors_header(path, file):
     return catalog, metadata, length_of_header + 8
 
     
-def torch_write_file(sd, file_path, quantization_map = None, config = None):
+def torch_write_file(sd, file_path, quantization_map = None, config = None, extra_meta = None):
     from collections import OrderedDict
     sf_sd = OrderedDict()
  
@@ -189,6 +217,14 @@ def torch_write_file(sd, file_path, quantization_map = None, config = None):
     if not config is None:
         metadata["config_base64"] = base64.b64encode(json.dumps(config, ensure_ascii=False).encode('utf8')).decode('utf8')
 
+    if not extra_meta is None:
+        for n , m in extra_meta.items():
+            if isinstance(m, str):
+                metadata[n] = m
+            else:
+                metadata[n + "_base64"] = base64.b64encode(json.dumps(m, ensure_ascii=False).encode('utf8')).decode('utf8')
+
+
     if len(metadata) > 0:
         sf_sd["__metadata__"] = metadata
 
@@ -214,7 +250,7 @@ def torch_write_file(sd, file_path, quantization_map = None, config = None):
                         t = t.view(torch.uint16)
                     elif  dtype ==  torch.float8_e5m2 or dtype ==  torch.float8_e4m3fn:
                         t = t.view(torch.uint8)
-                    buffer = t.numpy().tobytes()
+                    buffer = t.cpu().numpy().tobytes()
                     bytes_written = writer.write(buffer)            
                     assert bytes_written == size                    
             i+=1
@@ -224,7 +260,7 @@ def torch_write_file(sd, file_path, quantization_map = None, config = None):
 class SafeTensorFile:
     """Main class for accessing safetensors files that provides memory-efficient access"""
     
-    def __init__(self, file_path, metadata, catalog, skip_bytes, lazy_loading = True):
+    def __init__(self, file_path, metadata, catalog, skip_bytes, lazy_loading = True, writable_tensors = True):
         self._file_path = file_path
         self._metadata = metadata
         self._catalog = catalog
@@ -233,19 +269,20 @@ class SafeTensorFile:
         self.sd = None
         self.mtracker = None
         self.lazy_loading = lazy_loading
+        self.writable_tensors = writable_tensors
 
     @classmethod
-    def load_metadata(cls, file_path, lazy_loading = True):
+    def load_metadata(cls, file_path, lazy_loading = True, writable_tensors = True):
         with open(file_path, 'rb') as f:
             catalog, metadata, skip_bytes = _read_safetensors_header(file_path, f)
 
-        return cls(file_path, metadata, catalog, skip_bytes, lazy_loading)
+        return cls(file_path, metadata, catalog, skip_bytes, lazy_loading, writable_tensors )
 
-    def init_tensors(self, lazyTensors = True):
+    def init_tensors(self, lazyTensors = True, writable_tensors = True):
         if self.sd is None:
             self.lazy_loading = lazyTensors
             if lazyTensors:
-                self.sd = self.create_tensors_with_mmap()
+                self.sd = self.create_tensors_with_mmap(writable_tensors)
             else:
                 self.sd = self.create_tensors_without_mmap()
         # else:
@@ -255,7 +292,7 @@ class SafeTensorFile:
         return self.sd
     
             
-    def create_tensors_with_mmap(self):
+    def create_tensors_with_mmap(self, writable_tensors = True):
  
         self.mtracker = MmapTracker(self._file_path)
         import mmap
@@ -294,7 +331,7 @@ class SafeTensorFile:
         with open(self._file_path, 'rb') as f:
             i = 0
             for map_start, map_size in maps_info:
-                mm = mmap.mmap(f.fileno(), map_size, offset=map_start, access=mmap.ACCESS_COPY) #.ACCESS_READ
+                mm = mmap.mmap(f.fileno(), map_size, offset=map_start, access=  mmap.ACCESS_COPY  if writable_tensors else mmap.ACCESS_READ) 
                 maps.append((mm, map_start, map_size))
                 self.mtracker.register(mm, i, map_start, map_size)
                 i = i+ 1
@@ -348,10 +385,13 @@ class SafeTensorFile:
                 sd[k] = t
         return sd
 
+    def get_slice(self, name: str) -> torch.tensor:
+        return tensor_slice(self._catalog, name, self.get_tensor(name))
+    
     def get_tensor(self, name: str) -> torch.tensor:
         """Get a tensor by name"""
         # To do : switch to a JIT tensor creation per tensor
-        self.init_tensors()
+        self.init_tensors(self.lazy_loading, writable_tensors= self.writable_tensors)
         return self.sd[name]
  
     def keys(self) -> List[str]:
@@ -366,7 +406,7 @@ class SafeTensorFile:
         
     def tensors(self) -> Dict[str, torch.tensor]:
         """Get dictionary of all tensors"""
-        self.init_tensors(self.lazy_loading)
+        self.init_tensors(self.lazy_loading, writable_tensors= self.writable_tensors)
         return self.sd
         
     def metadata(self) -> Optional[Dict[str, str]]:
@@ -375,7 +415,7 @@ class SafeTensorFile:
         
     def __len__(self) -> int:
         """Get number of tensors"""
-        self.init_tensors(self.lazy_loading)
+        self.init_tensors(self.lazy_loading, writable_tensors= self.writable_tensors)
         return len(self.keys())
         
     def __contains__(self, key: str) -> bool:
@@ -393,17 +433,22 @@ class SafeTensorFile:
 class _SafeTensorLoader:
     """Context manager for loading SafeTensorFile"""
     
-    def __init__(self, filename: str ):
+    def __init__(self, filename: str, writable_tensors = True ):
         self.filename = Path(filename)
+        self.writable_tensors = writable_tensors
         self.sft = None
         if not self.filename.exists():
             raise FileNotFoundError(f"File not found: {filename}")
             
     def __enter__(self) -> SafeTensorFile:
         """Open file and return SafeTensorFile instance"""
-        
+        writable_tensors = self.writable_tensors
+
+        if all_tensors_are_read_only:
+            writable_tensors = False
+
         try:
-            self.sft = SafeTensorFile.load_metadata(self.filename)
+            self.sft = SafeTensorFile.load_metadata(self.filename, writable_tensors= writable_tensors)
             return self.sft 
             
         except Exception as e:
@@ -413,21 +458,31 @@ class _SafeTensorLoader:
     def __exit__(self, exc_type, exc_val, exc_tb) -> None:
         """Clean up resources"""        
         self.close()
-        
+
+    def get_tensor(self, name):
+        if self.sft == None:
+            self.__enter__()
+        return self.sft.get_tensor(name)
+
+    def get_slice(self, name):
+        if self.sft == None:
+            self.__enter__()
+        return self.sft.get_slice(name)
+
     def close(self) -> None:
         if self.sft != None:
             self.sft._free_resources()
         pass
 
 
-def safe_open(filename: str, framework: str = "pt",device = "cpu") -> _SafeTensorLoader:
+def safe_open(filename: str, framework: str = "pt",device = "cpu", writable_tensors = True) -> _SafeTensorLoader:
     if device != "cpu" or framework !="pt":
         return _old_safe_open(filename =filename, framework=framework, device=device)
-    return _SafeTensorLoader(filename)
+    return _SafeTensorLoader(filename, writable_tensors = writable_tensors)
 
-def torch_load_file( filename, device = 'cpu' ) -> Dict[str, torch.Tensor]:
+def torch_load_file( filename, device = 'cpu', writable_tensors = True) -> Dict[str, torch.Tensor]:
     sd = {}
-    with safe_open(filename, framework="pt", device = device ) as f:
+    with safe_open(filename, framework="pt", device = device, writable_tensors =writable_tensors  ) as f:
         for k in f.keys():
             sd[k] = f.get_tensor(k)
         return sd
@@ -443,6 +498,4 @@ try:
     transformers.modeling_utils.safe_open = safe_open
     transformers.modeling_utils.safe_load_file = torch_load_file
 except:
-    pass
-
-
+    pass
\ No newline at end of file
